from abc import abstractmethod
import logging


__all__ = ["ConversationalQueryRewriter"]


class ConversationalQueryRewriter:
    """Base conversational query reformulation class"""
    def __init__(self, name: str, verbose: bool = False):
        self.name = name
        self.turn_id: int = -1
        self.total_latency: float = 0.0
        self.verbose: bool = verbose

    @abstractmethod
    def rewrite(self, query: str) -> str:
        """Rewrite original query text"""
        raise NotImplementedError

    def reset_history(self):
        """Reset conversation history for model"""
        if self.verbose and self.turn_id > -1:
            turns = self.turn_id + 1
            logging.info(
                "Resetting {} after {} turns (average reformulation latency {:.4f}s)".format(
                    self.name, turns, self.total_latency / (turns)
                )
            )
        self.turn_id = -1
        self.total_latency = 0



import logging

from chatty_goose.cqr import Hqe, Ntr
from chatty_goose.pipeline import RetrievalPipeline
from chatty_goose.settings import HqeSettings, NtrSettings
from chatty_goose.types import CqrType, PosFilter
from parlai.core.agents import Agent, register_agent
from pyserini.search import SimpleSearcher


@register_agent("ChattyGooseAgent")
class ChattyGooseAgent(Agent):
    @classmethod
    def add_cmdline_args(cls, parser, partial_opt = None):
        parser.add_argument('--name', type=str, default='CQR', help="The agent's name.")
        parser.add_argument('--cqr_type', type=str, default='fusion', help="hqe, t5, or fusion")
        parser.add_argument('--episode_done', type=str, default='[END]', help="end signal for interactive mode")
        parser.add_argument('--hits', type=int, default=50, help="number of hits to retrieve from searcher")

        # Pyserini
        parser.add_argument('--k1', default=0.82, help='BM25 k1 parameter')
        parser.add_argument('--b', default=0.68, help='BM25 b parameter')
        parser.add_argument('--from_prebuilt', type=str, default='cast2019', help="Pyserini prebuilt index")
       
        # T5
        parser.add_argument('--from_pretrained', type=str, default='castorini/t5-base-canard', help="Huggingface T5 checkpoint")
        
        # HQE
        parser.add_argument('--M', default=5, type=int, help='aggregate historcial queries for first stage (BM25) retrieval')
        parser.add_argument('--eta', default=10, type=float, help='QPP threshold for first stage (BM25) retrieval')
        parser.add_argument('--R_topic', default=4.5, type=float, help='topic keyword threshold for first stage (BM25) retrieval')
        parser.add_argument('--R_sub', default=3.5, type=float, help='subtopic keyword threshold for first stage (BM25) retrieval')
        parser.add_argument('--filter', default='pos', help='filter word method: no, pos, stp')
        parser.add_argument('--verbose', action='store_true')
        return parser

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.name = opt["name"]
        self.episode_done = opt["episode_done"]
        self.cqr_type = CqrType(opt["cqr_type"])

        # Initialize searcher
        searcher = SimpleSearcher.from_prebuilt_index(opt["from_prebuilt"])
        searcher.set_bm25(float(opt["k1"]), float(opt["b"]))

        # Initialize retrievers
        retrievers = []
        if self.cqr_type == CqrType.HQE or self.cqr_type == CqrType.FUSION:
            hqe_settings = HqeSettings(
                M=opt["M"],
                eta=opt["eta"],
                R_topic=opt["R_topic"],
                R_sub=opt["R_sub"],
                filter=PosFilter(opt["filter"]),
                verbose=opt["verbose"],
            )
            hqe = Hqe(searcher, hqe_settings)
            retrievers.append(hqe)
        if self.cqr_type == CqrType.T5 or self.cqr_type == CqrType.FUSION:
            t5_settings = NtrSettings(model_name=opt["from_pretrained"], verbose=opt["verbose"])
            t5 = Ntr(t5_settings)
            retrievers.append(t5)

        self.rp = RetrievalPipeline(searcher, retrievers, int(opt["hits"]))

    def observe(self, observation):
        # Gather the last word from the other user's input
        self.query = observation.get("text", "")
        if observation.get("episode_done") or self.query == self.episode_done:
            logging.info("Resetting agent history")
            self.rp.reset_history()

    def act(self):
        if self.query == self.episode_done:
            return {"id": self.id, "text": "Session finished"}

        # Retrieve hits
        hits = self.rp.retrieve(self.query)
        if len(hits) == 0:
            result = "Sorry, I couldn't find any results"
        else:
            result = hits[0].raw
        return { "id": self.id, "text": result }


if __name__ == "__main__":
    from parlai.scripts.interactive import Interactive

    Interactive.main(model="ChattyGooseAgent", cqr_type="fusion")



import logging
from typing import List

from chatty_goose.cqr import ConversationalQueryRewriter
from chatty_goose.util import reciprocal_rank_fusion
from pygaggle.rerank.base import Query, Reranker, hits_to_texts
from pyserini.search import JSimpleSearcherResult, SimpleSearcher

__all__ = ["RetrievalPipeline"]


class RetrievalPipeline:
    """
    End-to-end conversational passage retrieval pipeline

    Parameters:
        searcher (SimpleSearcher): Pyserini searcher for Lucene index
        reformulators (List[ConversationalQueryRewriter]): List of CQR methods to use for first-stage retrieval
        searcher_num_hits (int): number of hits returned by searcher - default 10
        early_fusion (bool): flag to perform fusion before second-stage retrieval - default True
        reranker (Reranker): optional reranker for second-stage retrieval
        reranker_query_index (int): retriever index to use for reranking query - defaults to last retriever
        reranker_query_reformulator (ConversationalQueryRewriter): CQR method for generating reranker query,
                                                                   overrides reranker_query_index if provided
    """

    def __init__(
        self,
        searcher: SimpleSearcher,
        reformulators: List[ConversationalQueryRewriter],
        searcher_num_hits: int = 10,
        early_fusion: bool = True,
        reranker: Reranker = None,
        reranker_query_index: int = -1,
        reranker_query_reformulator: ConversationalQueryRewriter = None,
    ):
        self.searcher = searcher
        self.reformulators = reformulators
        self.searcher_num_hits = int(searcher_num_hits)
        self.early_fusion = early_fusion
        self.reranker = reranker
        self.reranker_query_index = reranker_query_index
        self.reranker_query_reformulator = reranker_query_reformulator

    def retrieve(self, query) -> List[JSimpleSearcherResult]:
        cqr_hits = []
        cqr_queries = []
        for cqr in self.reformulators:
            new_query = cqr.rewrite(query)
            hits = self.searcher.search(new_query, k=self.searcher_num_hits)
            cqr_hits.append(hits)
            cqr_queries.append(new_query)

        # Merge results from multiple CQR methods if required
        if self.early_fusion or self.reranker is None:
            cqr_hits = reciprocal_rank_fusion(cqr_hits)

        # Return results if no reranker
        if self.reranker is None:
            return cqr_hits

        # Get query for reranker
        if self.reranker_query_reformulator is None:
            rerank_query = cqr_queries[self.reranker_query_index]
        else:
            rerank_query = self.reranker_query_reformulator.rewrite(query)

        # Rerank results
        if self.early_fusion:
            results = self.rerank(
                rerank_query, cqr_hits[:self.searcher_num_hits])
        else:
            # Rerank all CQR results and fuse together
            results = []
            for hits in cqr_hits:
                results = self.rerank(rerank_query, hits)
            results = reciprocal_rank_fusion(results)
        return results

    def rerank(self, query, hits):
        if self.reranker is None:
            logging.info("Reranker not available, skipping reranking")
            return hits

        reranked = self.reranker.rerank(Query(query), hits_to_texts(hits))
        reranked_scores = [r.score for r in reranked]

        # Reorder hits with reranker scores
        reranked = list(zip(hits, reranked_scores))
        reranked.sort(key=lambda x: x[1], reverse=True)
        reranked_hits = [r[0] for r in reranked]
        return reranked_hits

    def reset_history(self):
        for cqr in self.reformulators:
            cqr.reset_history()

        if self.reranker_query_reformulator:
            self.reranker_query_reformulator.reset_history()



import logging
import time
import torch

from chatty_goose.settings import NtrSettings
from spacy.lang.en import English
from transformers import T5ForConditionalGeneration, T5Tokenizer

from .cqr import ConversationalQueryRewriter

__all__ = ["Ntr"]


class Ntr(ConversationalQueryRewriter):
    """Neural Transfer Reformulation using a trained T5 model"""

    def __init__(self, settings: NtrSettings = NtrSettings(), device: str = None):
        super().__init__("Ntr", verbose=settings.verbose)

        # Model settings
        self.max_length = settings.max_length
        self.num_beams = settings.num_beams
        self.early_stopping = settings.early_stopping

        device = device or ("cuda" if torch.cuda.is_available() else "cpu")
        self.device = torch.device(device)

        if self.verbose:
            logging.info(f"Initializing T5 using model {settings.model_name}...")
        self.model = (
            T5ForConditionalGeneration.from_pretrained(settings.model_name)
            .to(device)
            .eval()
        )
        self.tokenizer = T5Tokenizer.from_pretrained(settings.model_name)
        self.nlp = English()
        self.history = []

    def rewrite(self, query: str) -> str:
        start_time = time.time()
        self.turn_id += 1

        # Build input sequence from query and history
        self.history += [query]
        src_text = " ||| ".join(self.history)
        src_text = " ".join([tok.text for tok in self.nlp(src_text)])
        input_ids = self.tokenizer(
            src_text, return_tensors="pt", add_special_tokens=True
        ).input_ids.to(self.device)

        # Generate new sequence
        output_ids = self.model.generate(
            input_ids,
            max_length=self.max_length,
            num_beams=self.num_beams,
            early_stopping=self.early_stopping,
        )

        # Decode output
        rewrite_text = self.tokenizer.decode(
            output_ids[0, 0:],
            clean_up_tokenization_spaces=True,
            skip_special_tokens=True,
        )
        self.total_latency += time.time() - start_time
        return rewrite_text

    def reset_history(self):
        super().reset_history()
        self.history = []



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

excluded = ["data*", "examples*", "experiments*"]


setuptools.setup(
    name="chatty-goose",
    version="0.2.0",
    author="Anserini Gaggle",
    author_email="anserini.gaggle@gmail.com",
    description="A conversational passage retrieval toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/castorini/chatty-goose",
    install_requires=requirements,
    packages=setuptools.find_packages(exclude=excluded),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)
